from setuptools import setup, find_packages

setup(
    name                          = "ekosis",
    version                       = "0.11.0",
    url                           = "https://github.com/TheLastCylon/ecosystem",
    author                        = "Dirk Botha",
    author_email                  = "bothadj@gmail.com",
    description                   = "A Python framework for creating message-based, distributed systems. Allowing for TCP, UDP and UDS communications. Using JSON as the communications protocol.",
    long_description              = open('README.md').read(),
    packages                      = find_packages(include=['ekosis*']),
    long_description_content_type = 'text/markdown',
    install_requires              = ["sqlalchemy>=2.0.31", "pydantic>=2.8.2", "numpy>=2.0.0"],
    python_requires               = ">=3.11.4",
    license                       = "BSD-3-Clause",
    classifiers                   = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.11',
    ]
)



import uuid

from typing import Any, Dict, cast
from pydantic import BaseModel as PydanticBaseModel

from ..requests.endpoint import endpoint
from ..state_keepers.queued_sender_keeper import QueuedSenderKeeper
from ..util.utility_functions import string_to_uuid
from ..data_transfer_objects.queue_management import (
    QManagementRequestDto,
    QManagementItemRequestDto,
    QDatabaseSizesDto,
    QueuedSenderInformationDto,
    QManagementResponseDto
)

# --------------------------------------------------------------------------------
def make_queued_sender_information_dto(route_key: str, queue_info_dict: Dict[str, Any]) -> QueuedSenderInformationDto:
    return QueuedSenderInformationDto(
        route_key            = route_key,
        send_process_paused  = queue_info_dict["send_process_paused"],
        database_sizes       = QDatabaseSizesDto(
            pending = queue_info_dict["pending"],
            error   = queue_info_dict["error"]
        )
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_sender.data", QManagementRequestDto)
async def eco_queued_sender_data(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                 = cast(QManagementRequestDto, request)
    queued_sender_keeper = QueuedSenderKeeper()
    queue_info_dict      = await queued_sender_keeper.get_queue_information(data.queue_route_key)
    if not queue_info_dict:
        return QManagementResponseDto(message=f"No queued sender with route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data   = make_queued_sender_information_dto(data.queue_route_key, queue_info_dict),
        request_data = None,
        message      = f"Queued Sender[{data.queue_route_key}]: Data retrieved."
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_sender.send_process.pause", QManagementRequestDto)
async def eco_queued_sender_send_process_pause(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                 = cast(QManagementRequestDto, request)
    queued_sender_keeper = QueuedSenderKeeper()
    queue_info_dict      = await queued_sender_keeper.pause_send_process(data.queue_route_key)
    if not queue_info_dict:
        return QManagementResponseDto(message=f"No queued sender with route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data = make_queued_sender_information_dto(data.queue_route_key, queue_info_dict),
        message    = f"Queued Sender[{data.queue_route_key}]: Paused SEND PROCESS."
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_sender.send_process.unpause", QManagementRequestDto)
async def eco_queued_sender_send_process_unpause(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                 = cast(QManagementRequestDto, request)
    queued_sender_keeper = QueuedSenderKeeper()
    queue_info_dict      = await queued_sender_keeper.un_pause_send_process(data.queue_route_key)
    if not queue_info_dict:
        return QManagementResponseDto(message=f"No queued sender with route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data = make_queued_sender_information_dto(data.queue_route_key, queue_info_dict),
        message    = f"Queued Sender[{data.queue_route_key}]: UN-Paused SEND PROCESS."
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_sender.errors.get_first_10", QManagementRequestDto)
async def eco_queued_sender_errors_get_first_10(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                 = cast(QManagementRequestDto, request)
    queued_sender_keeper = QueuedSenderKeeper()
    uuid_list            = await queued_sender_keeper.get_first_x_error_uuids(data.queue_route_key)
    if uuid_list is None:
        return QManagementResponseDto(message=f"No queued sender with route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data = uuid_list,
        message    = f"Queued Sender[{data.queue_route_key}]: Retrieved first 10 UUIDs for error database."
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_sender.errors.reprocess.all", QManagementRequestDto)
async def eco_queued_sender_errors_reprocess_all(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                 = cast(QManagementRequestDto, request)
    queued_sender_keeper = QueuedSenderKeeper()
    queue_info_dict      = await queued_sender_keeper.reprocess_error_queue(data.queue_route_key)
    if not queue_info_dict:
        return QManagementResponseDto(message=f"No queued sender with route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data = make_queued_sender_information_dto(data.queue_route_key, queue_info_dict),
        message    = f"Queued Sender[{data.queue_route_key}]: Error database entries, moved to Retry database."
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_sender.errors.clear", QManagementRequestDto)
async def eco_queued_sender_errors_clear(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                 = cast(QManagementRequestDto, request)
    queued_sender_keeper = QueuedSenderKeeper()
    queue_info_dict      = await queued_sender_keeper.clear_error_queue(data.queue_route_key)
    if not queue_info_dict:
        return QManagementResponseDto(message=f"No queued sender with route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data = make_queued_sender_information_dto(data.queue_route_key, queue_info_dict),
        message    = f"Queued Sender[{data.queue_route_key}]: Error database cleared."
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_sender.errors.reprocess.one", QManagementItemRequestDto)
async def eco_queued_sender_errors_reprocess_one(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                 = cast(QManagementItemRequestDto, request)
    queued_sender_keeper = QueuedSenderKeeper()
    request_uid          = string_to_uuid(data.request_uid)
    if not request_uid:
        return QManagementResponseDto(message=f"[{data.request_uid}] is not a valid UUID.")

    queue_info_dict = await queued_sender_keeper.reprocess_error_queue_request_uid(data.queue_route_key, uuid.UUID(data.request_uid))

    if queue_info_dict is None:
        return QManagementResponseDto(message=f"No queued sender with route key: [{data.queue_route_key}]")
    elif not queue_info_dict:
        return QManagementResponseDto(message=f"No request with uid [{data.request_uid}] in queued sender error database, for route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data   = make_queued_sender_information_dto(data.queue_route_key, queue_info_dict),
        request_data = queue_info_dict["request_data"],
        message      = f"Queued Sender[{data.queue_route_key}]: Error database entry[{data.request_uid}] moved to Retry database."
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_sender.errors.pop_request", QManagementItemRequestDto)
async def eco_queued_sender_errors_pop_request(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                 = cast(QManagementItemRequestDto, request)
    queued_sender_keeper = QueuedSenderKeeper()
    request_uid          = string_to_uuid(data.request_uid)
    if not request_uid:
        return QManagementResponseDto(message=f"[{data.request_uid}] is not a valid UUID.")

    queue_info_dict = await queued_sender_keeper.pop_request_from_error_queue(data.queue_route_key, request_uid)
    if queue_info_dict is None:
        return QManagementResponseDto(message=f"No queued sender with route key: [{data.queue_route_key}]")
    elif not queue_info_dict:
        return QManagementResponseDto(message=f"No request with uid [{data.request_uid}] in queued sender error database, for route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data   = make_queued_sender_information_dto(data.queue_route_key, queue_info_dict),
        request_data = queue_info_dict["request_data"],
        message      = f"Queued Sender[{data.queue_route_key}]: POPPED Error database entry[{data.request_uid}]."
    )

# --------------------------------------------------------------------------------
@endpoint("eco.queued_sender.errors.inspect_request", QManagementItemRequestDto)
async def eco_queued_sender_errors_inspect_request(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    data                 = cast(QManagementItemRequestDto, request)
    queued_sender_keeper = QueuedSenderKeeper()
    request_uid          = string_to_uuid(data.request_uid)
    if not request_uid:
        return QManagementResponseDto(message=f"[{data.request_uid}] is not a valid UUID.")

    queue_info_dict = await queued_sender_keeper.inspect_request_in_error_queue(data.queue_route_key, uuid.UUID(data.request_uid))
    if queue_info_dict is None:
        return QManagementResponseDto(message=f"No queued sender with route key: [{data.queue_route_key}]")
    elif not queue_info_dict:
        return QManagementResponseDto(message=f"No request with uid [{data.request_uid}] in queued sender error database, for route key: [{data.queue_route_key}]")

    return QManagementResponseDto(
        queue_data   = make_queued_sender_information_dto(data.queue_route_key, queue_info_dict),
        request_data = queue_info_dict["request_data"],
        message      = f"Queued Sender[{data.queue_route_key}]: INSPECTING Error database entry[{data.request_uid}]."
    )



import asyncio
import logging

from ..server_base import ServerBase

from ...configuration.config_models import ConfigUDP

log = logging.getLogger()

# --------------------------------------------------------------------------------
class DatagramProtocolServer(asyncio.DatagramProtocol):
    __ENQ_byte    : int   =  5 # Decimal  5 = Ascii ENQ (enquiry) character
    __ACK_byte    : int   =  6 # Decimal  6 = Ascii ACK (acknowledge) character
    __LF_byte     : int   = 10 # Decimal 10 = Ascii LF (line feed) character = '\n'
    __ACK_response: bytes = bytes([__ACK_byte, __LF_byte])

    def __init__(self, build_response_function):
        self.build_response_function = build_response_function

        self.transport     : asyncio.DatagramTransport = None
        self.loop          : asyncio.AbstractEventLoop = None
        self.__write_lock  : asyncio.Lock              = asyncio.Lock()

    def connection_made(self, transport):
        self.transport = transport
        self.loop      = asyncio.get_running_loop()

    async def do_response(self, message, addr):
        async with self.__write_lock:
            response = await self.build_response_function(message)
            self.transport.sendto(response.encode(), addr)

    # TODO: find a way to deal with partial bytes read
    def datagram_received(self, bytes_read, addr):
        if bytes_read and bytes_read[-1] == self.__LF_byte:
            if bytes_read[0] == self.__ACK_byte: # The client wants to know we are here.
                self.transport.sendto(self.__ACK_response, addr)
            else:
                request_data: str = bytes_read.decode()
                self.loop.create_task(self.do_response(request_data, addr))

# --------------------------------------------------------------------------------
class UDPServer(ServerBase):
    def __init__(self, configuration : ConfigUDP):
        ServerBase.__init__(self)
        self.host       : str                       = configuration.host
        self.port       : int                       = configuration.port
        self.__transport: asyncio.DatagramTransport = None
        self.__loop     : asyncio.AbstractEventLoop = None

    # --------------------------------------------------------------------------------
    async def __aenter__(self):
        self._running = True
        self._logger.info(f"Setting up UDP server.")

    # --------------------------------------------------------------------------------
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    # --------------------------------------------------------------------------------
    def stop(self):
        if self._running:
            self._running = False
            self._logger.info(f"Stopping UDP server for {self.host}:{self.port}.")

    # --------------------------------------------------------------------------------
    async def __process_received_data(self, received_data: str):
        response_dict = await self._route_request(received_data)
        return response_dict.model_dump_json()

    # --------------------------------------------------------------------------------
    async def __create_datagram_listener(self):
        self._logger.info(f"Serving UDP on [{self.host}:{self.port}]")
        self.__loop = asyncio.get_running_loop()
        self.__transport, protocol = await self.__loop.create_datagram_endpoint(
            lambda: DatagramProtocolServer(self.__process_received_data),
            local_addr=(self.host, self.port)
        )

        try:
            while self._running:
                await asyncio.sleep(1)
        except asyncio.exceptions.CancelledError:
            self._logger.info("UDP server cancelled.")
        finally:
            self.__transport.close()

    # --------------------------------------------------------------------------------
    async def serve(self):
        if self._running:
            await self.__create_datagram_listener()



import uuid

from typing import Dict, List, cast
from pydantic import BaseModel as PydanticBaseModel

from .handler_base import HandlerBase
from .queued_handler_base import QueuedRequestHandlerBase
from .status import Status

from ..data_transfer_objects import RequestDTO
from ..util import SingletonType
from ..state_keepers.statistics_keeper import StatisticsKeeper

# --------------------------------------------------------------------------------
class RoutingExceptionBase(Exception):
    def __init__(self, status: int, message: str):
        self.status : int = status
        self.message: str = f"status: [{status}] message: [{message}]"
        super().__init__(self.message)

# --------------------------------------------------------------------------------
class UnknownRouteKeyException(RoutingExceptionBase):
    def __init__(self, route_key: str):
        super().__init__(Status.ROUTE_KEY_UNKNOWN.value, f"Unknown route key '{route_key}'")

# --------------------------------------------------------------------------------
class RequestRouter(metaclass=SingletonType):
    __statistics_keeper: StatisticsKeeper       = StatisticsKeeper()
    __routing_table    : Dict[str, HandlerBase] = {}

    def register_handler(self, handler: HandlerBase):
        if handler.get_route_key() not in self.__routing_table:
            self.__routing_table[handler.get_route_key()] = handler
            self.__statistics_keeper.track_endpoint_data(handler.get_route_key())
        else:
            raise Exception(f"Handler command id [{handler.get_route_key()}] already exists")

    def get_queued_handlers(self):
        response: List[QueuedRequestHandlerBase] = []
        for queue in self.__routing_table.values():
            if isinstance(queue, QueuedRequestHandlerBase):
                response.append(cast(QueuedRequestHandlerBase, queue))
        return response

    async def route_request(self, request: RequestDTO) -> PydanticBaseModel:
        if request.route_key not in self.__routing_table.keys():
            raise UnknownRouteKeyException(request.route_key)

        return await self.__routing_table[request.route_key].attempt_request(uuid.UUID(request.uid), request.data)

# --------------------------------------------------------------------------------
if __name__ == "__main__":
    print("Not an executable script, intended for use in other scripts")



import json

from typing import List

from ...data_transfer_objects.queue_management import (
    QManagementRequestDto,
    QManagementItemRequestDto,
    QManagementResponseDto
)

from ...sending.sender_base import SenderBase
from ...clients import ClientBase


# --------------------------------------------------------------------------------
def print_response(response: QManagementResponseDto):
    print("================================================================================")
    print(f"{response.message}")
    print("--------------------------------------------------------------------------------")
    if isinstance(response.queue_data, List):
        print(f"First 10 error UUIDs:\n{json.dumps(response.queue_data, indent=4)}")
    elif response.queue_data is not None:
        print(f"Queue information:\n{response.queue_data.model_dump_json(indent=4)}")
    if response.request_data:
        print("--------------------------------------------------------------------------------")
        print(f"Request information:\n{json.dumps(response.request_data, indent=4)}")
    print("================================================================================")


# --------------------------------------------------------------------------------
class QueueManagementSenderBase(SenderBase[QManagementRequestDto, QManagementResponseDto]):
    def __init__(self, client: ClientBase, route_key: str):
        super().__init__(
            client,
            route_key,
            QManagementRequestDto,
            QManagementResponseDto
        )

    async def send(self, queue_route_key: str):
        request_data = QManagementRequestDto(queue_route_key=queue_route_key)
        response = await self.send_data(request_data)
        print_response(response)


# --------------------------------------------------------------------------------
class QueueManagementItemSenderBase(SenderBase[QManagementItemRequestDto, QManagementResponseDto]):
    def __init__(self, client: ClientBase, route_key: str):
        super().__init__(
            client,
            route_key,
            QManagementItemRequestDto,
            QManagementResponseDto
        )

    async def send(self, queue_route_key: str, request_uid: str):
        request_data = QManagementItemRequestDto(
            queue_route_key = queue_route_key,
            request_uid     = request_uid
        )
        response = await self.send_data(request_data)
        print_response(response)



import uuid

from typing import Dict, List, Any, cast
from pydantic import BaseModel as PydanticBaseModel

from ..requests.endpoint import endpoint
from ..state_keepers.error_state_list import ErrorStateList


# --------------------------------------------------------------------------------
class ErrorsResponseDto(PydanticBaseModel):
    errors: List[Dict[str, Any]]


# --------------------------------------------------------------------------------
class ErrorCleanerRequestDto(PydanticBaseModel):
    error_id: str
    count   : int


# --------------------------------------------------------------------------------
def build_errors_response() -> List[Dict[str, Any]]:
    error_state_list = ErrorStateList()
    errors: List[Dict[str, Any]] = []
    for error_state in error_state_list.get_error_states():
        if error_state.is_set():
            errors.append(error_state.to_dict())
    return errors


# --------------------------------------------------------------------------------
@endpoint("eco.error_states.get")
async def eco_error_states_get(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    return ErrorsResponseDto(errors=build_errors_response())


# --------------------------------------------------------------------------------
@endpoint("eco.error_states.clear")
async def eco_error_states_clear(request_uuid: uuid.UUID, request) -> PydanticBaseModel:
    request          = cast(ErrorCleanerRequestDto, request)
    error_state_list = ErrorStateList()
    if request.count == 0:
        error_state_list.clear_all(request.error_id)
    else:
        error_state_list.clear_some(request.error_id, request.count)
    return ErrorsResponseDto(errors=build_errors_response())


# --------------------------------------------------------------------------------
if __name__ == "__main__":
    print("Not an executable script, intended for use in other scripts")



import asyncio
import logging

from .client_base import ClientBase

from ..exceptions import (
    CommunicationsNonRetryable,
    CommunicationsMaxRetriesReached,
)

log = logging.getLogger()

# --------------------------------------------------------------------------------
class DatagramProtocolClient(asyncio.DatagramProtocol):
    def __init__(self):
        self.response : asyncio.Future            = None
        self.transport: asyncio.DatagramTransport = None

    # --------------------------------------------------------------------------------
    def connection_made(self, transport):
        self.transport = transport

    # --------------------------------------------------------------------------------
    def datagram_received(self, data: bytes, address: tuple[str, int]) -> None:
        if self.response:
            self.response.set_result(data.decode())
            self.response = None

    # --------------------------------------------------------------------------------
    def send_message(self, message: str):
        if self.transport is not None:
            self.response = asyncio.Future()
            self.transport.sendto(message.encode())
            return self.response
        else:
            return None

# --------------------------------------------------------------------------------
class DatagramClientBase(ClientBase, asyncio.DatagramProtocol):
    def __init__(
        self,
        server_host: str,
        server_port: int,
        timeout    : float = 5,
        max_retries: int   = 3,
        retry_delay: float = 0.1,
    ):
        super().__init__(max_retries, retry_delay)
        self.server_host: str                       = server_host
        self.server_port: int                       = server_port
        self.timeout    : float                     = timeout
        self.initialised: bool                      = False
        self.loop       : asyncio.AbstractEventLoop = None
        self.transport  : asyncio.DatagramTransport = None
        self.protocol   : DatagramProtocolClient    = None
        self.send_lock  : asyncio.Lock              = asyncio.Lock()

    # --------------------------------------------------------------------------------
    async def _send_message(self, message: str) -> str:
        if not self.initialised:
            self.loop            = asyncio.get_running_loop()
            self.transport, self.protocol = await self.loop.create_datagram_endpoint(
                lambda: DatagramProtocolClient(),
                remote_addr=(self.server_host, self.server_port)
            )
            self.initialised = True

        async with self.send_lock:
            return await self.protocol.send_message(message)

    # --------------------------------------------------------------------------------
    async def _send_message_retry_loop(self, request: str) -> str:
        retry_count = 0
        while retry_count < self.max_retries:
            try:
                response = await self._send_message(request)
                return response
            except (TimeoutError, asyncio.TimeoutError) as e:
                retry_count += 1
                if retry_count >= self.max_retries:
                    raise CommunicationsMaxRetriesReached()
                else:
                    await asyncio.sleep(self.retry_delay)
            except Exception as e:
                raise CommunicationsNonRetryable(str(e))



from typing import Type, TypeVar
from pydantic import BaseModel as PydanticBaseModel

from .queued_sender_class import QueuedSenderClass

from ..clients import ClientBase
from ..data_transfer_objects import EmptyDto
from ..state_keepers.queued_sender_keeper import QueuedSenderKeeper

_RequestDTOType  = TypeVar("_RequestDTOType" , bound=PydanticBaseModel)
_ResponseDTOType = TypeVar("_ResponseDTOType", bound=PydanticBaseModel)


# --------------------------------------------------------------------------------
def queued_sender(
    client           : ClientBase,
    route_key        : str,
    request_dto_type : Type[_RequestDTOType],
    response_dto_type: Type[_ResponseDTOType] = EmptyDto,
    wait_period      : float                  = 0,
    page_size        : int                    = 100,
    max_retries      : int                    = 0,
):
    def inner_decorator(function):
        queued_sender_keeper   = QueuedSenderKeeper()
        queued_sender_instance = QueuedSenderClass[_RequestDTOType, _ResponseDTOType](
            client,
            route_key,
            request_dto_type,
            response_dto_type,
            wait_period,
            page_size,
            max_retries
        )
        queued_sender_keeper.add_queued_sender(queued_sender_instance)

        async def wrapper(*args, **kwargs):
            request_uid_to_use = None
            if "request_uid" in kwargs.keys():
                request_uid_to_use = kwargs["request_uid"]

            await queued_sender_instance.push_message(await function(*args, **kwargs), request_uid_to_use)
        return wrapper
    return inner_decorator
